import os
import numpy as np
import matplotlib.pyplot as plt
import ast
import MESAPC._MESAPC.utils as utils
import time


def load_dict(file):
    with open(file, 'r') as file:
        data = file.read().replace('\n', '')
    return ast.literal_eval(data)


class DataHandler():
    def __init__(self, inlist = "inlist_MESAPC"):
        basics, grid, runs = utils.read_file(inlist)
        self.basics = basics
        self.output_dir = basics["output_dir"]
        self.grid = None
        self.runs = []
        self.run_infos = runs
        self.num_runs = len(runs)
        self.load_data()

    def load_data(self):
        if self.basics["calculate_grid"]:
            self.grid = data_grid("grid", self.output_dir)
        for j in range(1, self.num_runs +1):
            self.runs.append(data_grid(f"run{j}", self.output_dir))

    def load_hists(self):
        if self.basics["calculate_grid"]:
            self.grid.load_hists()
        for run in self.runs:
            run.load_hists()

    def refresh_files(self):
        if self.basics["calculate_grid"]:
            self.grid.refresh_files()
        for run in self.runs:
            run.refresh_files()


class data_grid():
    def __init__(self, grid, output_dir):
        self.output_dir = output_dir
        self.grid_name = grid
        files = os.listdir(output_dir + '/LOGS')
        self.grid_files = [i for i in files if grid in i]
        self.grid_len = len(self.grid_files)
        self.data = []
        for i in range(1, self.grid_len + 1):
            self.data.append(evol_data(self.output_dir + '/LOGS/' + f"{grid}_{i}/"))

    def refresh_files(self):
        files = os.listdir(self.output_dir + '/LOGS')
        grid_files = [i for i in files if self.grid_name in i]
        for i in range(self.grid_len+ 1, len(grid_files) + 1):
            self.data.append(evol_data(self.output_dir + '/LOGS/' + f"{self.grid_name}_{i}/"))
        self.grid_files = grid_files
        self.grid_len = len(grid_files)
        return


    def load_hists(self):
        for evol in self.data:
            evol.load_hist()
        return

    def load_profiles(self):
        for evol in self.data:
            evol.load_profiles()
        return

    def plot_grid(self, ax, x, y, lw = 2, ls = '-', color = 'black', alpha = 1):
        for evol in self.data:
            if evol.hist_load_time != 0:
                ax.plot(evol.hist[x], evol.hist[y], lw = lw, ls =ls, color = color, alpha = alpha)
        return


class evol_data():
    def __init__(self, directory):
        self.controls =load_dict(directory + "controls_dict.txt")
        self.star_job =load_dict(directory + "star_dict.txt")
        self.dir = directory
        self.hist = None
        self.hist_keys = None
        self.hist_load_time = 0
        self.profiles = {}
        self.profile_keys = None

    def load_hist(self):
        try:
            savetime = os.path.getmtime(self.dir+ "history.data")
        except FileNotFoundError:
            return
        if savetime > self.hist_load_time:
            self.hist_load_time = savetime
            self.hist = np.genfromtxt(self.dir+ "history.data", skip_header =5, names = True)
            self.hist_keys = self.hist.dtype.names
        return

    def load_profiles(self):
        profiles = np.loadtxt(self.dir+ "profiles.index", skiprows= 1).T
        for profile_number in profiles[2]:
            self.load_profile(int(profile_number))
        return

    def load_profile(self, profile_number):
        self.profiles[f"profile{profile_number}"] =  np.genfromtxt(self.dir+ f"profile{profile_number}.data", skip_header =5, names = True)
        if self.profile_keys == None:
            self.profile_keys = self.profiles[f"profile{profile_number}"].dtype.names
        return




import os
import sys
import MESAPC._MESAPC.MESAPC_grid as MESAPC_grid
import MESAPC._MESAPC.utils as utils
import shutil


class MESAPC():
    def __init__(self,  output_dir = None, grid = None, inlist = "default",  runs = None, NUM_THREADS = os.environ["OMP_NUM_THREADS"] ):
        self.controls_list = None
        self.star_list = None
        if output_dir == None and grid == None:
            try:
                basics, grid, runs = utils.read_file("inlist_MESAPC")
            except FileNotFoundError:
                sys.exit("FATAL: No inlist_MESAPC in current directory. If you are wokring with the MESAPC class, you will have to at least give the output dir and grid dictionary.")
            try:
                self.output_dir = basics["output_dir"] + "/"
            except:
                sys.exit("FATAL: output dir not specified in_MESAPC.")
            self.calculate_grid = basics["calculate_grid"]
            self.NUM_THREADS = basics["num_threads"]
            self.TOTAL_NUM_THREADS=basics["total_num_threads"]
            self.runs = runs
            self.inlist = basics["inlist"]
            self.work_dir = basics["work_dir"]
            self.grid = grid
        else:
            self.output_dir = output_dir + "/"
            self.NUM_THREADS = NUM_THREADS
            self.grid = grid
            self.runs = []
            if runs != None:
                for r in range(len(runs)):
                    self.runs.append(utils.run_info(runs[r], r+1))
            self.inlist = inlist
        try:
            os.mkdir(self.output_dir)
        except FileExistsError:
            sys.exit(
                "FATAL: output dir already exist!")
        os.mkdir(self.output_dir + "/LOGS/")
        os.environ["OMP_NUM_THREADS"] = str(self.NUM_THREADS)
        self.grid = self.get_grid(self.grid)
        self.filebase_dir = os.environ["MESAPC_DIR"] + "/filebase/"
        if self.work_dir == "default":
            self.work_dir = os.environ["MESAPC_DIR"] + "/filebase/work"
        if self.inlist == "default":
            self.inlist = self.filebase_dir + "inlist_project"
        shutil.copy(self.inlist, self.output_dir + "LOGS/")
        if output_dir == None and grid == None:
            shutil.copy("inlist_MESAPC", self.output_dir + "LOGS/")
        if self.calculate_grid:
            self.grid.run_all()
        else:
            for r in self.runs:
                self.grid.run(r)

    def default_lists(self, type):
        defaults = open(os.environ["MESA_DIR"] + f"/star/defaults/{type}.defaults", 'r')
        Lines = defaults.readlines()
        pvals = []
        for line in Lines:
            line = line.strip()
            if len(line) == 0:
                continue
            if line[0] != "!":
                line=line.split("(")[0].lower()
                pvals.append(line.split(" ")[0].lower())
        return pvals

    def split_dict(self, grid, run={}, output="array"):
        if self.controls_list == None:
            self.controls_list = self.default_lists("controls")
            self.star_list = self.default_lists("star_job")
        c_dict = {}
        c_names = []
        c_vals = []
        s_dict = {}
        s_names = []
        s_vals = []
        if len(run.keys()) == 0:
            for key in grid.keys():
                if key in self.controls_list:
                    c_dict[key] = grid[key]
                    c_names.append(key)
                    c_vals.append(grid[key])
                elif key in self.star_list:
                    s_dict[key] = grid[key]
                    s_names.append(key)
                    s_vals.append(grid[key])
                else:
                    sys.exit(f"FATAL: {key} is no possible parameter for MESA!")
        if output == "array":
            return c_names, c_vals, s_names, s_vals
        elif output == "dict":
            return c_dict, s_dict

    def get_grid(self, grid):
        c_names, c_vals, s_names, s_vals = self.split_dict(grid)
        return MESAPC_grid.MESAPC_grid(self, c_names, c_vals, s_names, s_vals)



import itertools
import multiprocessing as mp
import MESAPC._MESAPC.utils as utils
import tqdm


class MESAPC_grid():
    def __init__(self, parent, c_names, c_vals, s_names, s_vals):
        self.parent = parent
        self.control_names = c_names
        self.control_vals = c_vals
        self.star_names = s_names
        self.star_vals = s_vals

    def run(self, run = None):
        MESA_runs = []
        if run != None:
            vals_new  = []
            indices = []
            current_index = 0
            for val in self.control_vals:
                if isinstance(val[0], list):
                    indices_list = ()
                    for val2 in val:
                        vals_new.append(val2)
                        indices_list += (current_index, )
                        current_index += 1
                    indices.append(indices_list)
                else:
                    vals_new.append(val)
                    current_index += 1
            c_names, c_vals, s_names, s_vals = self.parent.split_dict(run.run_dict)
            names = list(itertools.chain(self.control_names, self.star_names, c_names, s_names))
            vals = list(itertools.product(*vals_new, *self.star_vals, *c_vals, *s_vals))
            vals_new = []
            for tuple in vals:
                new_tuple = ()
                current_index =0
                current_array = []
                current_tuple = 0
                for element in tuple:
                    try:
                        if current_index in indices[current_tuple]:
                            current_array.append(element)
                            if len(current_array) == len(indices[current_tuple]):
                                new_tuple += (current_array,)
                                current_array = []
                                current_tuple +=1
                        else:
                            new_tuple += (element,)
                    except IndexError:
                        new_tuple += (element,)
                    current_index +=1
                vals_new.append(new_tuple)
            vals = vals_new
        else:
            vals_new  = []
            indices = []
            current_index = 0
            for val in self.control_vals:
                if isinstance(val[0], list):
                    indices_list = ()
                    for val2 in val:
                        vals_new.append(val2)
                        indices_list += (current_index, )
                        current_index += 1
                    indices.append(indices_list)
                else:
                    vals_new.append(val)
                    current_index += 1
            names = list(itertools.chain(self.control_names, self.star_names))
            # vals = list(itertools.product(*self.control_vals, *self.star_vals))
            vals = list(itertools.product(*vals_new, *self.star_vals))
            vals_new = []
            for tuple in vals:
                new_tuple = ()
                current_index =0
                current_array = []
                current_tuple = 0
                for element in tuple:
                    try:
                        if current_index in indices[current_tuple]:
                            current_array.append(element)
                            if len(current_array) == len(indices[current_tuple]):
                                new_tuple += (current_array,)
                                current_array = []
                                current_tuple +=1
                        else:
                            new_tuple += (element,)
                    except IndexError:
                        new_tuple += (element,)
                    current_index +=1
                vals_new.append(new_tuple)
            vals = vals_new
        for i in range(len(vals)):
            d = {}
            for j in range(len(names)):
                d[names[j]] = vals[i][j]
            c, s = self.parent.split_dict(d, output="dict")
            if  run != None:
                r_id = utils.run_identification(f"run{run.run_number}_{i + 1}", c, s, f"run{run.run_number}_{i + 1}",
                                                self.parent.inlist, self.parent.filebase_dir, self.parent.output_dir, self.parent.NUM_THREADS, self.parent.work_dir)
            else:
                r_id = utils.run_identification(f"grid_{i+1}", c, s, f"grid_{i+1}", self.parent.inlist,
                                                self.parent.filebase_dir, self.parent.output_dir, self.parent.NUM_THREADS, self.parent.work_dir)
            MESA_runs.append(r_id)

        num = int(self.parent.TOTAL_NUM_THREADS/self.parent.NUM_THREADS)
        pool = mp.Pool(processes=num)
        for _ in tqdm.tqdm(pool.imap_unordered(utils.run_mesa, MESA_runs), total=len(MESA_runs), smoothing=0):
            pass

    def run_all(self):
        self.run()
        for r in self.parent.runs:
            self.run(r)


import f90nml
import shutil
import os
import numpy as np
import subprocess as sp


class cd:
    """
    Directory changer. can change the directory using the 'with' keyword, and returns to the previous path
    after leaving intendation. Example:
    with cd("some/path/to/go"): # changing dir
        foo()
        ...
        bar()
    #back to old dir
    """
    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def run_mesa(run_id):
    inlist = run_id.inlist
    output_dir = run_id.output_dir
    file = f90nml.read(inlist)
    shutil.copytree(run_id.work_dir, output_dir + run_id.dir)
    for key in run_id.controls_dict.keys():
        file["controls"][key]= run_id.controls_dict[key]
    for key in run_id.star_dict.keys():
        file["star_job"][key]= run_id.star_dict[key]
    file["controls"]["log_directory"] = "../LOGS/" + run_id.log_dir
    file.write(output_dir +  run_id.dir + "/inlist_project")
    os.mkdir(output_dir + "/LOGS/" + run_id.log_dir)
    f = open(output_dir + "/LOGS/" + run_id.log_dir + "/controls_dict.txt", "w")
    f.write(str(run_id.controls_dict))
    f.close()
    f = open(output_dir + "/LOGS/" + run_id.log_dir + "/star_dict.txt", "w")
    f.write(str(run_id.star_dict))
    f.close()
    with cd(output_dir + run_id.dir):
        sp.call("./clean", shell = True, stderr=sp.DEVNULL, stdout=sp.DEVNULL)
        sp.call("./mk", stderr=sp.DEVNULL, stdout=sp.DEVNULL)
        sp.call("./star", stderr=sp.DEVNULL, stdout=sp.DEVNULL)
    return


class run_identification():
    def __init__(self, dir, controls_dict, star_dict, log_dir, inlist, filebase_dir, output_dir, num_threads, work_dir):
        self.dir = dir
        self.controls_dict = controls_dict
        self.star_dict = star_dict
        self.log_dir = log_dir
        self.inlist = inlist
        self.filebase_dir = filebase_dir
        self.work_dir = work_dir
        self.output_dir = output_dir
        self.OMP_NUM_THREADS = num_threads

class run_info():
    def __init__(self, run_dict, run_number):
        self.run_dict = run_dict
        self.run_number = run_number

def read_values(string):
    if isinstance(string, list):
        return_list = []
        for input_element in range(len(string)):
            return_list.append(read_values(string[input_element]))
        return return_list
    input = string.split(",")
    if input[0] == "float":
        return [float(i) for i in input[1:]]
    elif input[0] == "int":
        return [int(i) for i in input[1:]]
    elif input[0] == "string":
        return input[1:]
    elif input[0] == "linspace":
        return np.linspace(float(input[1]), float(input[2]), int(input[3]))
    elif input[0] == "logspace":
        return 10**(np.linspace(np.log10(float(input[1])), np.log10(float(input[2])), int(input[3])))

def basics_default():
    b = {
        "num_threads": 2,
        "filebase" : os.environ["MESAPC_DIR"] + "/filebase/",
        "calculate_grid": True,
        "total_num_threads": 10,
        "inlist": "default",
        "work_dir": "default"
    }
    return b


def read_file(string):
    file = f90nml.read(string)
    b = basics_default()
    for key in file["basics"].keys():
        b[key] = file["basics"][key]
    grid = {}
    for key in file["grid"].keys():
        grid[key] = read_values(file["grid"][key])
    runs = []
    keys = list(file.keys())
    count = 1
    for r in range(len(keys)):
        if keys[r][0] == "r":
            run_dict = {}
            for key in file[keys[r]].keys():
                run_dict[key] = read_values(file[keys[r]][key])
            runs.append(run_info(run_dict, count))
            count +=1
    return b, grid, runs




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# with open("requirements.txt", "r") as fh:
#     requirements = fh.read().splitlines()

setuptools.setup(
    name="MESAPC", # Replace with your own username
    version="0.0.1",
    author="Thomas Steindl",
    author_email="thomas.steindl95@gmx.at",
    description="Python package to calculate any grid of models with MESA",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
)



import matplotlib.pyplot as plt
import MESAPC._MESAPC.data_handler as data_handler



handler = data_handler.DataHandler()

fig, ax = plt.subplots(figsize = (12,8))

handler.load_hists()
colors = ['r', 'b', 'g']
while True:
    run_count = 0
    handler.grid.plot_grid(ax, 'log_Teff', 'log_L')
    ax.plot(ax.set_xlim()[0],ax.set_ylim()[0], 'k-', label ='grid')
    for run in handler.runs:
        run.plot_grid(ax, 'log_Teff', 'log_L', color = colors[run_count])
        ax.plot(ax.set_xlim()[0],ax.set_ylim()[0], f'{colors[run_count]}-', label = run.grid_name)
        run_count += 1
    ax.invert_xaxis()
    ax.set_ylabel('log($L/L_\odot$)')
    ax.set_xlabel('log($T_{\\rm eff}$)')
    ax.legend()
    handler.refresh_files()
    handler.load_hists()
    plt.draw()
    plt.pause(1)
    ax.clear()

